"""Utility functions."""

import os
import sys
import random

from typing import Callable, Iterable, List, Optional, Tuple, TypeVar, Union


T = TypeVar("T")


def set_randseed_from_shell() -> Tuple[Union[None, int, Tuple[int, ...]]]:
    """
    Set the random seed from the command line arguments.

    The random seed is set by hashing the command line arguments and using the hash as the seed.

    Returns:
        The random state.
    """
    random.seed(" ".join(sys.argv[1:]), version=2)
    return random.getstate()


def dos2unix(s: str) -> str:
    """
    Convert DOS line endings to Unix line endings.

    Args:
        s(str): The string to convert.

    Returns:
        The string with DOS line endings converted to Unix line endings.
    """
    return s.replace("\r\n", "\n")


def dos2unix_file(file: str) -> None:
    """
    Convert DOS line endings to Unix line endings in a file.

    Args:
        file(str): The file to convert.
    """
    data: List[str] = []

    with open(file, "r", encoding="ascii") as f:
        for line in f:
            data.append(line.strip())

    while data and not data[-1]:
        data.pop()

    with open(file, "w", encoding="ascii", newline="\n") as f:
        for line in data:
            print(line, file=f)


def dos2unix_dir(
    directory: str = ".",
    recursive: bool = False,
    *,
    suffixs: Tuple[str, ...] = ("in", "out"),
    echo: bool = False,
) -> None:
    """
    Convert DOS line endings to Unix line endings in a directory.

    Args:
        directory(str): The directory to convert (default: ".").
        recursive(bool): Whether to convert recursively (default: False).
        suffixs(Tuple[str]): The suffixes of the files to convert (default: ("in", "out")).
        echo(bool): Whether to print the files that are converted (default: False).
    """
    for root, _, files in os.walk(directory):
        for file in files:
            if any(file.endswith(suffix) for suffix in suffixs):
                dos2unix_file(os.path.join(root, file))
                if echo:
                    print(f"Converted {ascii(file)} to Unix format.")
        if not recursive:
            break


def filter_none(iterable: Iterable[Optional[T]]) -> Iterable[T]:
    """Filter out the None values in an iterable."""
    return (x for x in iterable if x is not None)


def gener_uniform(start: int, end: int) -> Callable[..., int]:
    """
    Generate a uniform random number generator function.

    Args:
        start(int): The start of the range.
        end(int): The end of the range.

    Returns:
        A function that generates a random integer in the range [start, end).
    """

    def generator(
        *_args: int, **_kwargs: int
    ) -> int:  # pylint: disable=unused-argument
        """Generate a random integer in the range [start, end)."""
        return random.randrange(start, end)

    return generator


class Dsu:
    """Disjoint Set Union data structure."""

    def __init__(self, n: int):
        self.par = list(range(n))
        self.rank = [0] * n

    def find(self, x: int) -> int:
        """Finds the representative (root) of the set containing the element `x`."""
        if self.par[x] != x:
            self.par[x] = self.find(self.par[x])
        return self.par[x]

    def union(self, x: int, y: int) -> bool:
        """Unions the sets containing the elements `x` and `y`."""
        x, y = self.find(x), self.find(y)
        if x == y:
            return False
        if self.rank[x] < self.rank[y]:
            x, y = y, x
        if self.rank[x] == self.rank[y]:
            self.rank[x] += 1
        self.par[y] = x
        return True



"""
This module provides two classes, `Seq` and `SeqIter`, for working with mathematical sequences.
"""

import sys
import warnings

from typing import Any, Callable, Dict, Generator, Iterable, Optional, Union, overload


Number = Any


class SeqIter:
    """
    An iterator class for iterating over a sequence (`Seq`) with customizable
    start, stop, and step parameters.
    Attributes:
        _seq (Seq): The sequence to iterate over.
        start (Optional[int]): The starting index of the iteration. Defaults to 0.
        stop (Optional[int]): The stopping index of the iteration. Defaults to infinity.
        step (Optional[int]): The step size for the iteration. Defaults to 1.
    Methods:
        __iter__(): Returns the iterator object itself.
        __next__(): Returns the next value in the sequence based on the current
                    position and step size. Raises StopIteration when the end
                    of the iteration is reached.
    """

    def __init__(
        self,
        _seq: "Seq",
        start: Optional[int] = None,
        stop: Optional[int] = None,
        step: Optional[int] = None,
    ):
        self.__seq = _seq
        self.__start = start if start is not None else 0
        self.__stop = stop if stop is not None else float("inf")
        self.__step = step if step is not None else 1
        self.__pos = self.__start

    def __iter__(self):
        return self

    def __next__(self) -> Number:
        if self.__step >= 0 and self.__pos >= self.__stop:
            raise StopIteration
        if self.__step < 0 and self.__pos <= self.__stop:
            raise StopIteration
        value = self.__seq[self.__pos]
        self.__pos += self.__step
        return value


class Seq:
    """
    Seq is a class that represents a mathematical sequence, allowing for the calculation of sequence values
    based on a provided formula or a generator function. The generator function is used to avoid recursion.
    """

    __type: bool
    __iter_limit: Optional[int]
    __formula: Callable[[int, Callable[[int], Number]], Number]
    __yield_formula: Callable[[int], Generator[int, Number, Number]]
    __values: Dict[int, Number]

    def __init__(
        self,
        formula: Optional[Callable[[int, Callable[[int], Number]], Number]] = None,
        *values0: Number,
        values1: Optional[Dict[int, Number]] = None,
        yield_formula: Optional[Callable[[int], Generator[int, Number, Number]]] = None,
        iteration_limit: Optional[int] = int(1e6),
    ):
        """
        Create a new sequence object.

        Args:
            formula (Callable[[int, Callable[[int], Number]], Number]): The formula to calculate the sequence.
                The first argument is the index of the value to calculate.
                The second argument is a function to get the value of the sequence at a given index.
            yield_formula (Callable[[int], Generator[int, Number, Number]]): The formula to calculate the sequence.
                The argument is the index of the value to calculate.
                The generator should yield the indices that need to be calculated.
            *values0 (Number): The initial values of the sequence.
            values1 (Dict[int, Number]): The values of the sequence at the indices specified in this dictionary.
                These values will overwrite the values in *values0.

        Raises:
            ValueError: If neither formula nor yield_formula is provided.
            ValueError: If both formula and yield_formula are provided.

        Warnings:
            RuntimeWarning: If a value is overwritten in the values1 dictionary.

        Examples:
            >>> s = Seq(lambda i, f: f(i - 1) + f(i - 2), 0, 1)
            >>> s[0]
            0
            >>> s[1]
            1
            >>> s[2]
            1
            >>> list(s[:10])
            [0, 1, 1, 2, 3, 5, 8, 13, 21, 34]
            >>> def fib_gen(i):
            ...     if i < 2:
            ...         return i
            ...     x = yield i - 1
            ...     y = yield i - 2
            ...     return (x + y) % 998244353
            ...
            >>> s = Seq(yield_formula=fib_gen)
            >>> s[100000]
            10519474
        """
        if formula is not None:
            self.__type = False
            self.__formula = formula
        if yield_formula is not None:
            self.__type = True
            self.__yield_formula = yield_formula
        if formula is None and yield_formula is None:
            raise ValueError("Either formula or yield_formula must be provided")
        if formula is not None and yield_formula is not None:
            raise ValueError("Only one of formula and yield_formula can be provided")
        self.__iter_limit = iteration_limit

        self.__values = {} if values1 is None else values1.copy()
        for i, v in enumerate(values0):
            if i in self.__values:
                warnings.warn(
                    f"Duplicate value for index {i}, overwriting with new value",
                    RuntimeWarning,
                    2,
                )
            else:
                self.__values[i] = v

    @property
    def has_yield_formula(self) -> bool:
        """Whether the sequence has a yield formula."""
        return self.__type

    @property
    def iter_limit(self) -> Optional[int]:
        """The iteration limit of the sequence."""
        return self.__iter_limit

    def __calc_with_yield(self, i: int, gen: Generator[int, Number, Number]) -> Number:
        stack = [(i, gen)]
        ret: Optional[Number] = None
        while True:
            try:
                if ret is not None:
                    cur = stack[-1][1].send(ret)
                    ret = None
                else:
                    cur = next(stack[-1][1])
                if cur in self.__values:
                    ret = self.__values[cur]
                else:
                    stack.append((cur, self.__yield_formula(cur)))
                    if self.__iter_limit is not None and len(stack) > self.__iter_limit:
                        raise RecursionError("Iteration limit exceeded")
            except StopIteration as stop:
                d, _ = stack.pop()
                ret = self.__values[d] = stop.value
                if not stack:
                    return stop.value

    def __call__(self, i: int) -> Number:
        """
        Get the value of the sequence at index i.

        Args:
            i: The index of the value to get.

        Returns:
            The value of the sequence at index i.
        """
        if i not in self.__values:
            if not self.has_yield_formula:
                self.__values[i] = self.__formula(i, self)
            else:
                self.__calc_with_yield(i, self.__yield_formula(i))
        return self.__values[i]

    @overload
    def __getitem__(self, ind: int) -> Number: ...

    @overload
    def __getitem__(self, ind: slice) -> Iterable[Number]: ...

    def __getitem__(self, ind: Union[int, slice]) -> Union[Number, Iterable[Number]]:
        """
        Get the value of the sequence at index i.

        Args:
            i: The index of the value to get.

        Returns:
            The value of the sequence at index i.
        """
        if isinstance(ind, int):
            return self.__call__(ind)
        return SeqIter(self, ind.start, ind.stop, ind.step)

    def __setitem__(self, i: int, value: Number):
        """
        Set the value of the sequence at index i.

        Args:
            i: The index of the value to set.
            value: The value to set the sequence at index i to.
        """
        self.__values[i] = value

    def __iter__(self):
        return iter(self.__values.items())

    def getiterationlimit(self) -> Optional[int]:
        """
        Get the iteration limit of the sequence.

        Returns:
            The iteration limit of the sequence.
        """
        return self.__iter_limit

    def setiterationlimit(self, limit: Optional[int]):
        """
        Set the iteration limit of the sequence.

        Args:
            limit (Optional[int]): The new iteration limit. None for no limit.
        """
        self.__iter_limit = limit

    @staticmethod
    def getrecursionlimit() -> int:
        """
        Get the recursion limit of the sequence.

        Returns:
            The recursion limit of the sequence.
        """
        return sys.getrecursionlimit()

    @staticmethod
    def setrecursionlimit(limit: int):
        """
        Set the recursion limit of the sequence.

        Args:
            limit (int): The new recursion limit.
        """
        sys.setrecursionlimit(limit)



"""
This module provides functions to generate graphs with a given degree sequence.
"""

import math
import random
import warnings

from typing import (
    Any,
    Callable,
    Counter,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

from rdgc.graph.base import Graph


class SwitchGraph:
    """A graph which can switch edges quickly"""

    __directed: bool
    __edges: List[Tuple[int, int]]
    __edges_set: Counter[Tuple[int, int]]

    def __init__(
        self,
        directed: bool = False,
        edge_seq: Sequence[Tuple[int, int]] = (),
    ):
        self.__directed = directed
        self.__edges = []
        self.__edges_set = Counter()
        for u, v in edge_seq:
            self.insert(u, v)

    def count(self, u: int, v: int) -> bool:
        """
        Returns the number of edges between vertices `u` and `v`.
        """
        if not self.__directed and u > v:
            u, v = v, u
        return (u, v) in self.__edges_set

    def insert(self, u: int, v: int) -> None:
        """
        Inserts an edge between vertices `u` and `v`.
        """
        if not self.__directed and u > v:
            u, v = v, u
        self.__edges.append((u, v))
        self.__edges_set[(u, v)] += 1

    def remove(self, ind: Sequence[int]) -> None:
        """
        Removes edges at the given indices by swapping them to the end of the list and then removing.

        Args:
            ind (Sequence[int]): The indices of the edges to remove.
        """
        for t, i in enumerate(sorted(ind, reverse=True), start=1):
            assert 0 <= i < len(self.__edges)
            self.__edges[i], self.__edges[-t] = self.__edges[-t], self.__edges[i]
        for _ in range(len(ind)):
            u, v = self.__edges.pop()
            self.__edges_set[(u, v)] -= 1
            if self.__edges_set[(u, v)] == 0:
                self.__edges_set.pop((u, v))

    def _switch3(
        self,
        first: int,
        second: int,
        *,
        self_loop: bool = False,
        multiedge: bool = False,
    ) -> bool:
        """
        Switches three random edges in the graph.

        Returns:
            bool: True if the edges are switched, otherwise False.
        """
        if len(self.__edges) < 3:
            return False
        u0, u1 = self.__edges[first]
        v0, v1 = self.__edges[second]
        assert u0 == u1 and v0 == v1

        third = random.choice(range(len(self.__edges)))
        while third in (first, second):
            third = random.choice(range(len(self.__edges)))
        x, y = self.__edges[third]

        # u -> v v -> y x -> u
        if not self_loop:
            if u0 == v0 or v0 == y or x == u0:
                return False
        if {(u0, u1), (v0, v1), (x, y)} == {(u0, v0), (v0, y), (x, u0)}:
            return False
        if not multiedge and len({(u0, v0), (v0, y), (x, u0)}) < 3:
            return False

        self.remove([first, second, third])

        if self.count(u0, v0) or self.count(v0, y) or self.count(x, u0):
            self.insert(u0, u1)
            self.insert(v0, v1)
            self.insert(x, y)
            return False

        self.insert(u0, v0)
        self.insert(v0, y)
        self.insert(x, u0)
        return True

    def switch(self, *, self_loop: bool = False, multiedge: bool = False) -> bool:
        """
        Switches two or three random edges in the graph.

        Returns:
            bool: True if the edges are switched, otherwise False.
        """
        if len(self.__edges) < 2:
            return False

        first, second = random.choices(range(len(self.__edges)), k=2)

        x1, y1 = self.__edges[first]
        x2, y2 = self.__edges[second]

        if self_loop:
            if x1 == x2 or y1 == y2:
                return False
        else:
            if {x1, y1} & {x2, y2} != set():
                return False

        if not multiedge and (self.count(x1, y2) or self.count(x2, y1)):
            return False

        if not multiedge and (x1 == y1 and x2 == y2):
            return self._switch3(
                first, second, self_loop=self_loop, multiedge=multiedge
            )

        self.remove([first, second])

        self.insert(x1, y2)
        self.insert(x2, y1)

        return True

    @staticmethod
    def from_directed_degree_sequence(
        degree_sequence: Sequence[Tuple[int, int]],
        *,
        self_loop: bool = False,
        multiedge: bool = False,
    ) -> "SwitchGraph":
        """
        Returns a graph with the given directed degree sequence.

        Note the time complexity of this function is O(n*m*log(n)),
        where n is the number of vertices and m is the number of edges.

        Args:
            degree_sequence (Sequence[Tuple[int, int]]): The directed degree sequence.
            self_loop (bool, optional): Specifies whether self-loops are allowed. Defaults to False.
            multiedge (bool, optional): Specifies whether multiple edges are allowed. Defaults to False.

        Raises:
            ValueError: If the degree sequence is invalid.

        Returns:
            SwitchGraph: A graph with the given directed degree sequence.
        """
        if any(x < 0 or y < 0 for (x, y) in degree_sequence):
            raise ValueError("Degree sequence must be non-negative")

        x, y = zip(*degree_sequence)
        if sum(x) != sum(y):
            raise ValueError("Degree sequence is not graphical")

        ret = SwitchGraph(True)

        if len(degree_sequence) == 0:
            return ret

        degseq = [[sout, sin, vn] for vn, (sin, sout) in enumerate(degree_sequence)]
        degseq.sort(reverse=True)

        try:
            while max(s[1] for s in degseq) > 0:
                kk = [i for i in range(len(degseq)) if degseq[i][1] > 0]
                _, in_d, vto = degseq[kk[0]]
                degseq[kk[0]][1] = 0
                j = 0
                while in_d:
                    _, _, vfrom = degseq[j]
                    if vto == vfrom and not self_loop:
                        j += 1
                        _, _, vfrom = degseq[j]
                    while in_d and degseq[j][0]:
                        in_d -= 1
                        degseq[j][0] -= 1
                        ret.insert(vfrom, vto)
                        if not multiedge:
                            break
                    j += 1
                degseq.sort(reverse=True)
        except IndexError as err:
            raise ValueError("Degree sequence is not graphical") from err

        return ret

    @staticmethod
    def from_undirected_degree_sequence(
        degree_sequence: Sequence[int],
        *,
        self_loop: bool = False,
        multiedge: bool = False,
    ) -> "SwitchGraph":
        """
        Returns a graph with the given undirected degree sequence.

        Note the time complexity of this function is O(n*m*log(n)),
        where n is the number of vertices and m is the number of edges.

        Args:
            degree_sequence (Sequence[int]): The undirected degree sequence.
            self_loop (bool, optional): Specifies whether self-loops are allowed. Defaults to False.
            multiedge (bool, optional): Specifies whether multiple edges are allowed. Defaults to False.

        Raises:
            ValueError: If the degree sequence is invalid.

        Returns:
            SwitchGraph: A graph with the given undirected degree sequence.
        """
        if any(x < 0 for x in degree_sequence):
            raise ValueError("Degree sequence must be non-negative")

        if sum(degree_sequence) % 2 != 0:
            raise ValueError("Degree sequence is not graphical")

        if len(degree_sequence) == 0:
            return SwitchGraph(False)

        degseq = [[d, vn] for vn, d in enumerate(degree_sequence)]
        degseq.sort(reverse=True)

        edges: List[Tuple[int, int]] = []
        try:
            while len(edges) * 2 < sum(degree_sequence):
                deg, x = degseq[0]
                degseq[0][0] = 0
                if self_loop:
                    while deg > 1:
                        deg -= 2
                        edges.append((x, x))
                        if not multiedge:
                            break
                y = 1
                while deg:
                    while deg and degseq[y][0]:
                        deg -= 1
                        degseq[y][0] -= 1
                        edges.append((x, degseq[y][1]))
                        if not multiedge:
                            break
                    y += 1
                degseq.sort(reverse=True)
        except IndexError as err:
            raise ValueError("Degree sequence is not graphical") from err

        return SwitchGraph(False, edges)

    @staticmethod
    def k_regular(
        size: int,
        k: int,
        *,
        self_loop: bool = False,
        multiedge: bool = False,
    ) -> "SwitchGraph":
        """
        Returns a k-regular graph with `size` vertices and degree `k`.

        Args:
            size (int): The number of vertices.
            k (int): The degree of each vertex.
            self_loop (bool, optional): Specifies whether self-loops are allowed. Defaults to False.
            multiedge (bool, optional): Specifies whether multiple edges are allowed. Defaults to False.

        Raises:
            ValueError: If the degree is invalid.

        Returns:
            SwitchGraph: A k-regular graph with `size` vertices and degree `k`.
        """
        if k < 0:
            raise ValueError("Degree must be non-negative")
        if size * k % 2 != 0:
            raise ValueError("Degree sequence is not graphical")
        if not multiedge:
            max_deg = size if self_loop else size - 1
            if k > max_deg:
                raise ValueError(f"Degree is too large: {k} > {max_deg}")
        if size == 0 or k == 0:
            return SwitchGraph()
        sg = SwitchGraph()
        if k % 2 == 1:
            # size % 2 == 0
            half = size // 2
            for i in range(half):
                sg.insert(i, (i + half) % size)
        k //= 2
        start = 0 if self_loop else 1
        for i in range(size):
            for j in range(start, start + k):
                sg.insert(i, (i + j) % size)
        return sg

    def edge_count(self) -> int:
        """Returns the total number of edges in the graph."""
        return len(self.__edges)

    def __iter__(self) -> Iterator[Tuple[int, int]]:
        return iter(self.__edges)


def from_degree_sequence(
    degree_sequence: Union[Sequence[int], Sequence[Tuple[int, int]]],
    iter_times: Optional[int] = None,
    *args: Any,
    self_loop: bool = False,
    multiedge: bool = False,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    iter_limit: int = int(1e6),
    **kwargs: Any,
) -> Graph:
    """
    Returns a graph with the given degree sequence.

    Note the time complexity of this function is O(n*m*log(n)),
    where n is the number of vertices and m is the number of edges.

    Args:
        degree_sequence (Union[Sequence[int], Sequence[Tuple[int, int]]]): The degree sequence.
        iter_times (int, optional): The number of iterations. Defaults to None.
        self_loop (bool, optional): Specifies whether self-loops are allowed. Defaults to False.
        multiedge (bool, optional): Specifies whether multiple edges are allowed. Defaults to False.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.
        iter_limit (int, optional): The maximum number of iterations. Defaults to 1e6.

    Raises:
        ValueError: If the degree sequence is invalid.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A graph with the given degree sequence.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    if weight_gener is None:
        weight_gener = lambda u, v: None

    if len(degree_sequence) == 0:
        return Graph(0, False)

    directed = not isinstance(degree_sequence[0], int)
    if directed:
        sg = SwitchGraph.from_directed_degree_sequence(
            cast(Sequence[Tuple[int, int]], degree_sequence),
            self_loop=self_loop,
            multiedge=multiedge,
        )
    else:
        sg = SwitchGraph.from_undirected_degree_sequence(
            cast(Sequence[int], degree_sequence),
            self_loop=self_loop,
            multiedge=multiedge,
        )

    size = len(degree_sequence)
    edge_count = sg.edge_count()
    if iter_times is None:
        iter_times = int(
            Graph.edge_estimate(size, edge_count, directed, self_loop, multiedge)
            / math.log(2)
        )
    iter_times = min(iter_times + 1, iter_limit)

    for _ in range(iter_times):
        sg.switch(self_loop=self_loop, multiedge=multiedge)
    graph = Graph(size, directed)
    for u, v in sg:
        graph.add_edge(u, v, weight_gener(u, v))
    return graph


def k_regular(
    size: int,
    k: int,
    iter_times: Optional[int] = None,
    *args: Any,
    self_loop: bool = False,
    multiedge: bool = False,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    iter_limit: int = int(1e6),
    **kwargs: Any,
) -> Graph:
    """
    Returns a k-regular graph with `size` vertices and degree `k`.

    Args:
        size (int): The number of vertices.
        k (int): The degree of each vertex.
        iter_times (int, optional): The number of iterations. Defaults to None.
        self_loop (bool, optional): Specifies whether self-loops are allowed. Defaults to False.
        multiedge (bool, optional): Specifies whether multiple edges are allowed. Defaults to False.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.
        iter_limit (int, optional): The maximum number of iterations. Defaults to 1e6.

    Raises:
        ValueError: If the degree is invalid.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A k-regular graph with `size` vertices and degree `k`.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    if weight_gener is None:
        weight_gener = lambda u, v: None

    sg = SwitchGraph.k_regular(size, k, self_loop=self_loop, multiedge=multiedge)
    edge_count = sg.edge_count()
    if iter_times is None:
        iter_times = int(
            Graph.edge_estimate(size, edge_count, False, self_loop, multiedge)
            / math.log(2)
        )
    iter_times = min(iter_times + 1, iter_limit)

    for _ in range(iter_times):
        sg.switch(self_loop=self_loop, multiedge=multiedge)
    graph = Graph(size, False)
    for u, v in sg:
        graph.add_edge(u, v, weight_gener(u, v))
    return graph



"""RDGC Data Generator for Contests"""

from rdgc.nt import (
    prime_sieve,
    prime_sieve2,
    is_prime,
    miller_rabin_test,
    miller_rabin,
    fast_isPrime,
    nextprime,
    prevprime,
    randprime,
)
from rdgc.seq import Seq, SeqIter, Number
from rdgc.graph import Graph, GRAPH_GENERS, make_graph
from rdgc.utils import (
    set_randseed_from_shell,
    dos2unix,
    dos2unix_file,
    dos2unix_dir,
    filter_none,
    gener_uniform,
)


__all__ = [
    "prime_sieve",
    "prime_sieve2",
    "is_prime",
    "miller_rabin_test",
    "miller_rabin",
    "fast_isPrime",
    "nextprime",
    "prevprime",
    "randprime",
    "Seq",
    "SeqIter",
    "Number",
    "Graph",
    "make_graph",
    "GRAPH_GENERS",
    "set_randseed_from_shell",
    "dos2unix",
    "dos2unix_file",
    "dos2unix_dir",
    "filter_none",
    "gener_uniform",
]



"""
A module to represent a graph.
"""

import itertools
import math
import random
import warnings
from collections import defaultdict

from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
)

from rdgc.utils import filter_none


class Graph:
    """A class to represent a graph."""

    __directed: bool
    __total_edge_count: int
    __vertex_ranks: List[Any]
    __adjacency_list: List[List[Any]]
    __indeg_counter: List[int]
    __outdeg_counter: List[int]
    __edge_counter: Dict[Tuple[int, int], int]

    def __init__(self, vertices: int, directed: bool = False, rnk: Iterable[Any] = ()):
        """
        Initializes a graph object.

        Args:
            vertices (int): The number of vertices in the graph. Must be non-negative.
            directed (bool, optional): Specifies whether the graph is directed. Defaults to False.
            edge_count (bool, optional): If True, enables tracking of edge counts. Defaults to False.

        Raises:
            ValueError: If the number of vertices is negative.

        Attributes:
            __directed (bool): Indicates if the graph is directed.
            __edge_cnt (int): Tracks the number of edges in the graph.
            __edges (list): Adjacency list representation of the graph.
            __edge_set (defaultdict or None): Tracks edge counts if `edge_count` is True, otherwise None.
        """
        if vertices < 0:
            raise ValueError("Number of vertices must be non-negative")
        self.__directed = directed
        self.__total_edge_count = 0
        self.__vertex_ranks = [None for _ in range(vertices)]
        self.__adjacency_list = [[] for _ in range(vertices)]
        self.__indeg_counter = [0] * vertices if directed else []
        self.__outdeg_counter = [0] * vertices
        self.__edge_counter = defaultdict(int)
        self.set_ranks(rnk)

    @property
    def edges(self) -> int:
        """Returns the total number of edges in the graph."""
        return self.__total_edge_count

    @property
    def vertices(self) -> int:
        """Returns the total number of vertices in the graph."""
        return len(self.__adjacency_list)

    @property
    def directed(self) -> bool:
        """Returns True if the graph is directed, otherwise False."""
        return self.__directed

    def __add_edge(self, u: int, v: int) -> None:
        """Increments the edge count between vertices `u` and `v`."""
        self.__edge_counter[(u, v)] += 1
        if not self.__directed and u != v:
            self.__edge_counter[(v, u)] += 1

    def set_rank(self, u: int, rnk: Any = None):
        """
        Sets the rank of vertex `u`.

        Args:
            u (int): The vertex.
            rnk (Any, optional): The rank of the vertex. Defaults to None.

        Raises:
            IndexError: If the vertex is invalid.
        """
        self.__vertex_ranks[u] = rnk

    def get_rank(self, u: int) -> Any:
        """
        Returns the rank of vertex `u`.

        Args:
            u (int): The vertex.

        Returns:
            Any: The rank of the vertex.

        Raises:
            IndexError: If the vertex is invalid.
        """
        return self.__vertex_ranks[u]

    def set_ranks(self, rnk: Iterable[Any], default: Any = None) -> None:
        """
        Sets the ranks of vertices `u`.

        Args:
            rnk (Iterable[Any]): The ranks of the vertices.
            default (Any, optional): The default rank. Defaults to None.

        Raises:
            IndexError: If the vertex is invalid.
        """
        rnk = itertools.chain(rnk, itertools.repeat(default))
        rnk = itertools.islice(rnk, self.vertices)
        for i, r in enumerate(rnk):
            self.set_rank(i, r)

    def get_ranks(self) -> Tuple[Any]:
        """
        Returns the ranks of all vertices.

        Returns:
            List[Any]: The ranks of all vertices.
        """
        return tuple(self.__vertex_ranks)

    def indegree(self, u: int) -> int:
        """
        Returns the indegree of vertex `u`.

        If the graph is undirected, please use the `degree` method instead.

        Args:
            u (int): The vertex.

        Raises:
            IndexError: If the vertex is invalid.
            ValueError: If the graph is undirected.

        Returns:
            int: The indegree of vertex `u`.
        """
        if not self.__directed:
            raise ValueError("Cannot calculate indegree of an undirected graph")
        return self.__indeg_counter[u]

    def outdegree(self, u: int) -> int:
        """
        Returns the outdegree of vertex `u`.

        If the graph is undirected, please use the `degree` method instead.

        Args:
            u (int): The vertex.

        Raises:
            IndexError: If the vertex is invalid.
            ValueError: If the graph is undirected.

        Returns:
            int: The outdegree of vertex `u`.
        """
        if not self.__directed:
            raise ValueError("Cannot calculate outdegree of an undirected graph")
        return len(self.__adjacency_list[u])

    def degree(self, u: int) -> int:
        """
        Returns the degree of vertex `u`.

        If the graph is directed, please use the `indegree` and `outdegree` methods instead.

        Args:
            u (int): The vertex.

        Raises:
            IndexError: If the vertex is invalid.
            ValueError: If the graph is directed.

        Returns:
            int: The degree of vertex `u`.
        """
        if self.__directed:
            raise ValueError("Cannot calculate degree of a directed graph")
        return self.__outdeg_counter[u]

    def get_edges(
        self, u: Optional[int] = None, *, directed: bool = False
    ) -> Generator[Tuple[int, int, Any], None, None]:
        """
        Returns an iterator over the edges of the graph.

        Args:
            u (int, optional): If specified, returns edges incident to vertex `u`. Defaults to None.
            directed (bool): Only works if the graph is undirected.
                If True, returns directed edges. Defaults to False.

        Yields:
            Tuple[int, int, Any]: A tuple containing the vertices and weight of an edge.
        """
        directed = self.__directed or directed
        if u is None:
            for uu in range(self.vertices):
                for vv, weight in self.__adjacency_list[uu]:
                    if directed or uu <= vv:
                        yield uu, vv, weight
        else:
            for v, weight in self.__adjacency_list[u]:
                yield u, v, weight

    def add_edge(self, u: int, v: int, weight: Any = None) -> None:
        """
        Adds an edge between vertices `u` and `v` with an optional weight.

        Args:
            u (int): The source vertex.
            v (int): The destination vertex.
            weight (Any, optional): The weight of the edge. Defaults to None.

        Raises:
            ValueError: If the vertices are invalid.
        """
        if u not in range(self.vertices) or v not in range(self.vertices):
            raise ValueError("Invalid vertex")
        self.__adjacency_list[u].append((v, weight))
        if not self.__directed and u != v:
            self.__adjacency_list[v].append((u, weight))
        self.__add_edge(u, v)
        self.__outdeg_counter[u] += 1
        if not self.__directed:
            self.__outdeg_counter[v] += 1
        else:
            self.__indeg_counter[v] += 1
        self.__total_edge_count += 1

    def add_rand_edge(
        self,
        edge_count: int,
        *,
        self_loop: bool = False,
        multiedge: bool = False,
        weight_gener: Optional[Callable[[int, int], Any]] = None,
    ) -> None:
        """
        Adds `edge_count` random edges to the graph.

        Args:
            edge_count (int): The number of edges to add.
            self_loop (bool, optional): Specifies whether self-loops are allowed. Defaults to False.
            multiedge (bool, optional): Specifies whether multiple edges are allowed. Defaults to False.
            weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.

        Raises:
            ValueError: If the number of edges is invalid.
        """
        if weight_gener is None:
            weight_gener = lambda u, v: None
        if not multiedge:
            max_edge = Graph.calc_max_edge(self.vertices, self.directed, self_loop)
            if self.edges + edge_count > max_edge:
                raise ValueError(
                    f"Too many edges: {self.edges} + {edge_count} > {max_edge}"
                )

        while self.edges < edge_count:
            u, v = (
                random.sample(range(self.vertices), 2)
                if not self_loop
                else random.choices(range(self.vertices), k=2)
            )
            if multiedge or self.count_edge(u, v) == 0:
                self.add_edge(u, v, weight_gener(u, v))

    def count_edges(self, u: int) -> int:
        """
        Returns the number of edges incident to vertex `u`.

        Args:
            u (int): The vertex.

        Returns:
            int: The number of edges incident to vertex `u`.

        Raises:
            ValueError: If the vertex is invalid.
        """
        if u not in range(self.vertices):
            raise ValueError("Invalid vertex")
        return len(self.__adjacency_list[u])

    def count_edge(self, u: int, v: int) -> int:
        """
        Returns the number of edges between vertices `u` and `v`.

        Args:
            u (int): The source vertex.
            v (int): The destination vertex.

        Returns:
            int: The number of edges between vertices `u` and `v`.

        Raises:
            ValueError: If the vertices are invalid or edge count is not enabled.
        """
        if u not in range(self.vertices) or v not in range(self.vertices):
            raise ValueError("Invalid vertex")
        return 0 if (u, v) not in self.__edge_counter else self.__edge_counter[(u, v)]

    def shuffle_nodes(
        self, shuffle: Optional[Callable[[int], Sequence[int]]] = None
    ) -> "Graph":
        """
        Shuffles the vertices of the graph.

        Args:
            shuffle (Callable[[int], Sequence[int]], optional): A function to shuffle the vertices. Defaults to None.

        Returns:
            Graph: A new graph with shuffled vertices.
        """
        mapping = (
            random.sample(range(self.vertices), self.vertices)
            if shuffle is None
            else shuffle(self.vertices)
        )
        if shuffle is not None:
            if len(mapping) != self.vertices:
                raise ValueError("Invalid shuffle mapping")
            if len(set(mapping)) != self.vertices:
                warnings.warn(
                    "Duplicate vertices in shuffle mapping", RuntimeWarning, 2
                )
        new_graph = Graph(self.vertices, self.directed)
        for u in range(self.vertices):
            new_graph.set_rank(mapping[u], self.get_rank(u))
        for u, v, w in self.get_edges():
            new_graph.add_edge(mapping[u], mapping[v], w)
        return new_graph

    def copy(self) -> "Graph":
        """
        Returns a deep copy of the graph.

        Returns:
            Graph: A deep copy of the graph.
        """
        new_graph = Graph(self.vertices, self.directed)
        for u in range(self.vertices):
            new_graph.set_rank(u, self.get_rank(u))
        for u, v, w in self.get_edges():
            new_graph.add_edge(u, v, w)
        return new_graph

    def transpose(self) -> "Graph":
        """
        Returns the transpose of the graph.

        Returns:
            Graph: The transpose of the graph.
        """
        if not self.directed:
            raise ValueError("Cannot transpose an undirected graph")
        new_graph = Graph(self.vertices, self.directed)
        for u, v, w in self.get_edges():
            new_graph.add_edge(v, u, w)
        return new_graph

    def output_nodes(
        self,
        *,
        sep: str = " ",
        printer: Optional[Callable[[int], str]] = None,
    ) -> str:
        """
        Returns a string representation of the vertices.

        Returns:
            str: A string representation of the vertices.
        """
        return sep.join(map(str if printer is None else printer, self.get_ranks()))

    def output_edges(
        self,
        *,
        sep: str = "\n",
        shuffle: bool = False,
        printer: Optional[Callable[[int, int, Any], Optional[str]]] = None,
    ) -> str:
        """
        Returns a string representation of the graph.

        Args:
            shuffle (bool, optional): If True, shuffles the edges. Defaults to False.
            printer (Callable[[int, int, Any], str], optional): A function to format the edges. Defaults to `f"{u} {v} {w}"`.

        Returns:
            str: A string representation of the graph.
        """
        if printer is None:
            printer = lambda u, v, w: f"{u} {v}" if w is None else f"{u} {v} {w}"

        edge_output = [printer(u, v, w) for u, v, w in self.get_edges()]
        if shuffle:
            random.shuffle(edge_output)

        return sep.join(filter_none(edge_output))

    @staticmethod
    def calc_max_edge(size: int, directed: bool, self_loop: bool):
        """
        Calculates the maximum number of edges in a graph with `size` vertices.

        Args:
            size (int): The number of vertices.
            directed (bool): Specifies whether the graph is directed.
            self_loop (bool): Specifies whether self-loops are allowed.

        Returns:
            int: The maximum number of edges in the graph
        """
        max_edge = size * (size - 1)
        if not directed:
            max_edge //= 2
        if self_loop:
            max_edge += size
        return max_edge

    @staticmethod
    def _estimate_comb(n: int, k: int) -> float:
        try:
            return float(sum(math.log(n - i) - math.log(i + 1) for i in range(k)))
        except ValueError:
            return 0.0

    @staticmethod
    def edge_estimate(
        size: int,
        edge_count: int,
        directed: bool,
        self_loop: bool,
        multiedge: bool,
    ) -> float:
        """
        Estimates the number of graphs with `size` vertices and `edge_count` edges.

        Args:
            size (int): The number of vertices.
            edge_count (int): The number of edges.
            directed (bool): Specifies whether the graph is directed.
            self_loop (bool): Specifies whether self-loops are allowed.
            multiedge (bool): Specifies whether multiple edges are allowed.

        Returns:
            float: The estimated number of graphs.
        """
        tot_edge = Graph.calc_max_edge(size, directed, self_loop)
        if multiedge:
            return Graph._estimate_comb(edge_count + tot_edge - 1, edge_count)
        return Graph._estimate_comb(tot_edge, edge_count)



"""
This module provides utility functions to generate graphs.
"""

import itertools
import warnings
import random as rd

from typing import Any, Callable, Literal, Optional, Tuple

from rdgc.graph.base import Graph


def null(
    size: int,
    *args: Any,
    directed: bool = False,
    **kwargs: Any,
) -> Graph:
    """
    Returns a null graph with `size` vertices.

    Args:
        size (int): The number of vertices.
        directed (bool, optional): Specifies whether the graph is directed. Defaults to False.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A null graph with `size` vertices.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    return Graph(size, directed)


def complete(
    size: int,
    *args: Any,
    directed: bool = False,
    self_loop: bool = False,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    **kwargs: Any,
) -> Graph:
    """
    Returns a complete graph with `size` vertices.

    Args:
        size (int): The number of vertices.
        directed (bool, optional): Specifies whether the graph is directed. Defaults to False.
        self_loop (bool, optional): Specifies whether self-loops are allowed. Defaults to False.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A complete graph with `size` vertices.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    if weight_gener is None:
        weight_gener = lambda u, v: None

    graph = Graph(size, directed)
    for u in range(size):
        for v in range(size) if directed else range(u, size):
            if u != v or self_loop:
                graph.add_edge(u, v, weight_gener(u, v))
    return graph


def tournament(
    size: int,
    *args: Any,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    **kwargs: Any,
) -> Graph:
    """
    Returns a tournament graph with `size` vertices.

    Args:
        size (int): The number of vertices.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A tournament graph with `size` vertices.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    if weight_gener is None:
        weight_gener = lambda u, v: None

    graph = Graph(size, True)
    for u in range(size):
        for v in range(u + 1, size):
            i, j = rd.sample((u, v), 2)
            graph.add_edge(i, j, weight_gener(i, j))
    return graph


def random_graph(
    size: int,
    edge_count: int,
    *args: Any,
    directed: bool = False,
    self_loop: bool = False,
    multiedge: bool = False,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    **kwargs: Any,
) -> Graph:
    """
    Returns a random graph with `size` vertices and `edge_count` edges.

    Args:
        size (int): The number of vertices.
        edge_count (int): The number of edges.
        directed (bool, optional): Specifies whether the graph is directed. Defaults to False.
        self_loop (bool, optional): Specifies whether self-loops are allowed. Defaults to False.
        multiedge (bool, optional): Specifies whether multiple edges are allowed. Defaults to False.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.

    Raises:
        ValueError: If the number of edges is invalid.

    Warnings:
        RuntimeWarning: If extra arguments are provided

    Returns:
        Graph: A random graph with `size` vertices and `edge_count` edges.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    if weight_gener is None:
        weight_gener = lambda u, v: None

    if not multiedge:
        max_edge = Graph.calc_max_edge(size, directed, self_loop)
        if edge_count > max_edge:
            raise ValueError(f"Too many edges: {edge_count} > {max_edge}")

    graph = Graph(size, directed)
    while graph.edges < edge_count:
        u, v = (
            rd.sample(range(size), 2) if not self_loop else rd.choices(range(size), k=2)
        )
        if multiedge or graph.count_edge(u, v) == 0:
            graph.add_edge(u, v, weight_gener(u, v))
    return graph


def cycle(
    size: int,
    *args: Any,
    directed: bool = False,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    **kwargs: Any,
) -> "Graph":
    """
    Returns a cycle graph with `size` vertices.

    Args:
        size (int): The number of vertices.
        directed (bool, optional): Specifies whether the graph is directed. Defaults to False.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A cycle graph with `size` vertices.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    if weight_gener is None:
        weight_gener = lambda u, v: None

    graph = Graph(size, directed)
    for u in range(size):
        graph.add_edge(u, (u + 1) % size, weight_gener(u, (u + 1) % size))
    return graph


def wheel(
    size: int,
    *args: Any,
    directed: bool = False,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    **kwargs: Any,
) -> "Graph":
    """
    Returns a wheel graph with `size` vertices.

    Args:
        size (int): The number of vertices.
        directed (bool, optional): Specifies whether the graph is directed. Defaults to False.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A wheel graph with `size` vertices.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    if weight_gener is None:
        weight_gener = lambda u, v: None

    graph = Graph(size, directed)
    for u in range(1, size):
        graph.add_edge(0, u, weight_gener(0, u))
    for u in range(1, size):
        graph.add_edge(u, u % (size - 1) + 1, weight_gener(u, u % (size - 1) + 1))
    return graph


def union(
    *graphs: Graph,
    directed: bool = False,
    self_loop: bool = False,
    multiedge: bool = False,
    node_mapping: Optional[Callable[[int, int, Tuple[Graph, ...]], int]] = None,
    default_mapping: Literal["separate", "combine", "connect"] = "separate",
) -> Graph:
    """
    Returns the union of multiple graphs.

    This function combines multiple graphs into a single graph, allowing for
    custom node mapping to handle potential conflicts in node IDs.

    The rank of nodes is preserved from the original graphs, only the last one which is not None is used.

    If `multiedge` is False, it ensures that no multiple edges are created between the same pair of nodes.
    If there are multiple edges, only the first one is kept.

    If `node_mapping` is not provided, it defaults to mapping nodes from different graphs
    to unique IDs by adding the cumulative number of vertices from previous graphs.

    Args:
        *graphs (Graph): The graphs to be unioned.
        directed (bool, optional): Specifies whether the resulting graph is directed. Defaults to False.
        self_loop (bool, optional): If True, allows self-loops in the resulting graph. Defaults to False.
        multiedge (bool, optional): If True, allows multiple edges between the same pair of nodes. Defaults to False.
        node_mapping (Callable[[int, int, tuple[Graph, ...]], int], optional):
            A function to map nodes from the original graphs to the new graph. It takes three arguments.
            + `node`: The node ID in the original graph.
            + `graph_index`: The index of the graph in the input list.
            + `graphs`: The tuple of all input graphs.
        default_mapping (Literal["separate", "combine", "connect"]):
            Specifies the default behavior for `node_mapping` if not provided.
            + If `"separate"`, it maps nodes to unique IDs by adding the cumulative number of vertices from previous graphs.
            + If `"combine"`, it keeps the original node IDs as they are.
            + If `"connect"`, it maps nodes to unique IDs except the first and last node of each graph,
              which are connected to the previous graph's last node and the next graph's first node, respectively.

    Returns:
        Graph: A new graph that is the union of the input graphs.
    """

    if not graphs:
        return Graph(0, directed)

    if node_mapping is None:
        if default_mapping == "combine":
            node_mapping = lambda node, graph_idx, graphs: node
        elif default_mapping == "separate":
            prev_sum = list(
                itertools.accumulate([0] + [graph.vertices for graph in graphs])
            )
            node_mapping = (
                lambda node, graph_index, graphs: node + prev_sum[graph_index]
            )
        elif default_mapping == "connect":
            prev_sum = list(
                itertools.accumulate([0] + [graph.vertices for graph in graphs])
            )
            node_mapping = (
                lambda node, graph_index, graphs: node
                + prev_sum[graph_index]
                - graph_index
            )
        else:
            raise ValueError(
                f"Unknown default_mapping: {default_mapping}. "
                "Use 'separate' or 'combine'."
            )

    max_node_id = max(
        node_mapping(j, i, graphs)
        for i in range(len(graphs))
        for j in range(graphs[i].vertices)
    )

    g = Graph(max_node_id + 1, directed=directed)

    for i, graph in enumerate(graphs):
        for u, v, weight in graph.get_edges(directed=directed):
            new_u = node_mapping(u, i, graphs)
            new_v = node_mapping(v, i, graphs)

            if not multiedge and g.count_edge(new_u, new_v) > 0:
                continue
            if not self_loop and new_u == new_v:
                continue

            g.add_edge(new_u, new_v, weight)

        for u in range(graph.vertices):
            new_u = node_mapping(u, i, graphs)
            u_rnk = graph.get_rank(u)
            if u_rnk is not None:
                g.set_rank(new_u, u_rnk)

    return g



"""
This module contains functions to generate connected graphs.
"""

import random
import warnings

from typing import Any, Callable, Optional

from rdgc.graph.base import Graph
from rdgc.graph.tree import spanning_tree
from rdgc.graph.utils import cycle as cyc


def connected(
    size: int,
    edge_count: int,
    *args: Any,
    directed: bool = False,
    self_loop: bool = False,
    cycle: bool = False,
    multiedge: bool = False,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    **kwargs: Any,
) -> "Graph":
    """
    Returns a connected graph with `size` vertices and `edge_count` edges.

    Args:
        size (int): The number of vertices.
        edge_count (int): The number of edges.
        directed (bool, optional): Specifies whether the graph is directed. Defaults to False.
        self_loop (bool, optional): Specifies whether self-loops are allowed. Defaults to False.
        multiedge (bool, optional): Specifies whether multiple edges are allowed. Defaults to False.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.

    Raises:
        ValueError: If the number of edges is invalid.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A connected graph with `size` vertices and `edge_count` edges.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    if cycle and not directed:
        warnings.warn(
            "Argument `cycle` is ignored for undirected graphs", RuntimeWarning, 2
        )

    if edge_count < size - 1:
        raise ValueError(f"Too few edges: {edge_count} < {size - 1}")
    if not multiedge:
        max_edge = Graph.calc_max_edge(size, directed and cycle, self_loop)
        if edge_count > max_edge:
            raise ValueError(f"Too many edges: {edge_count} > {max_edge}")

    if weight_gener is None:
        weight_gener = lambda u, v: None

    graph = Graph(size, directed)
    for u, v, _ in spanning_tree(size).get_edges():
        u, v = sorted((u, v))
        graph.add_edge(u, v, weight_gener(u, v))
    while graph.edges < edge_count:
        u, v = (
            random.sample(range(size), 2)
            if not self_loop
            else random.choices(range(size), k=2)
        )
        if not cycle:
            u, v = sorted((u, v))
        if multiedge or graph.count_edge(u, v) == 0:
            graph.add_edge(u, v, weight_gener(u, v))
    return graph


def strongly_connected(
    size: int,
    edge_count: int,
    *args: Any,
    self_loop: bool = False,
    multiedge: bool = False,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    **kwargs: Any,
) -> "Graph":
    """
    Returns a strongly connected graph with `size` vertices and `edge_count` edges.

    Args:
        size (int): The number of vertices.
        edge_count (int): The number of edges.
        self_loop (bool, optional): Specifies whether self-loops are allowed. Defaults to False.
        multiedge (bool, optional): Specifies whether multiple edges are allowed. Defaults to False.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.

    Raises:
        ValueError: If the number of edges is invalid.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A strongly connected graph with `size` vertices and `edge_count` edges.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    if weight_gener is None:
        weight_gener = lambda u, v: None

    if edge_count < size:
        raise ValueError(f"Too few edges: {edge_count} < {size}")
    if not multiedge:
        max_edge = Graph.calc_max_edge(size, True, self_loop)
        if edge_count > max_edge:
            raise ValueError(f"Too many edges: {edge_count} > {max_edge}")

    graph = cyc(size, directed=True, weight_gener=weight_gener)

    while graph.edges < edge_count:
        u, v = (
            random.sample(range(size), 2)
            if not self_loop
            else random.choices(range(size), k=2)
        )
        if multiedge or graph.count_edge(u, v) == 0:
            graph.add_edge(u, v, weight_gener(u, v))

    return graph



"""
This module provides some commonly used number theory methods.
"""

import random

from typing import List, Tuple


def prime_sieve(n: int) -> List[int]:
    """
    Return a list of prime numbers less than or equal to n.

    The time complexity is O(n * log(log(n))).

    However, it is usually faster than the O(n) version.

    If n is float, it will be converted to int.

    Args:
        n(int): The upper bound.

    Returns:
        List[int]: A list of prime numbers less than or equal to n.
    """
    n = int(n)
    vis = [True] * (n + 1)
    for i in range(2, int(n**0.5) + 1):
        if vis[i]:
            for j in range(i * i, n + 1, i):
                vis[j] = False
    return [i for i in range(2, n + 1) if vis[i]]


def prime_sieve2(n: int) -> List[int]:
    """
    Return a list of prime numbers less than or equal to n.

    The time complexity is O(n).

    If n is float, it will be converted to int.

    Args:
        n(int): The upper bound.

    Returns:
        List[int]: A list of prime numbers less than or equal to n.
    """
    n = int(n)
    vis: List[bool] = [True] * (n + 1)
    prime: List[int] = []
    for i in range(2, n + 1):
        if vis[i]:
            prime.append(i)
        for p in prime:
            if p * i > n:
                break
            vis[p * i] = False
            if i % p == 0:
                break
    return prime


def is_prime(n: int) -> bool:
    """
    Check if n is a prime number.

    The time complexity is O(sqrt(n)).

    Args:
        n(int): The number to be checked.

    Returns:
        bool: True if n is a prime number, False otherwise.
    """
    if not isinstance(n, int):
        raise ValueError("n must be an integer")
    if n < 2:
        return False
    for i in range(2, int(n**0.5) + 1):
        if n % i == 0:
            return False
    return True


def miller_rabin_test(n: int, r: int, d: int, a: int) -> bool:
    """
    Perform the Miller-Rabin primality test for a given base.
    The Miller-Rabin test is a probabilistic algorithm used to determine
    whether a number is a probable prime. This function tests the primality
    of a number `n` using a specific base `a`.

    Args:
        n (int): The number to test for primality.
        r (int): The number of times `n - 1` can be divided by 2 (exponent of 2 in `n - 1`).
        d (int): The odd part of `n - 1` such that `n - 1 = 2^r * d`.
        a (int): The base to use for the test.

    Returns:
        bool: True if `n` passes the Miller-Rabin test for the given base `a`,
              indicating that `n` is a probable prime. False otherwise.
    """
    x = pow(a, d, n)
    if x in (0, 1, n - 1):
        return True
    for _ in range(r - 1):
        x = x * x % n
        if x == n - 1:
            return True
    return False


def miller_rabin(n: int, k: int = 16, *, bases: Tuple[int, ...] = ()) -> bool:
    """
    Check if a number is a prime using the Miller-Rabin primality test.

    The Miller-Rabin algorithm is a probabilistic test to determine if a number is a prime.
    It is efficient and widely used for large numbers.

    Note that the algorithm is probabilistic, meaning that it can return false positives;
    however, the probability of a false positive can be reduced by increasing the number of iterations `k`.

    The algorithm has a time complexity of O(k * log(n)). For most applications, `k = 16` is sufficient.

    Args:
        n (int): The number to be tested for primality.
        k (int, optional): The number of iterations to perform. Higher values of `k`
            increase the accuracy of the test. Defaults to 16.

    Returns:
        bool: Returns `True` if the number `n` is likely a prime, and `False` otherwise.

    Raises:
        ValueError: If `n` is not an integer.
    """
    if not isinstance(n, int):
        raise ValueError("n must be an integer")
    if n < 2:
        return False
    if n in [2, 3, 5, 7, 11, 13, 17, 19]:
        return True
    if any(n % p == 0 for p in [2, 3, 5, 7, 11, 13, 17, 19]):
        return False
    r, d = 0, n - 1
    while d % 2 == 0:
        r += 1
        d //= 2
    for b in bases:
        if not miller_rabin_test(n, r, d, b):
            return False
    for _ in range(k - len(bases)):
        if not miller_rabin_test(n, r, d, random.randint(2, n - 2)):
            return False
    return True


fast_isPrime = miller_rabin


def nextprime(n: int) -> int:
    """
    Find the next prime number greater than or equal to the given integer `n`.

    Notes:
        - If `n` is less than 2, the function returns 2 as the smallest prime number.
        - The function uses the Miller-Rabin primality test to check for prime numbers.
        - The algorithm optimizes the search by leveraging properties of numbers
          in the 6k ± 1 form, where k is an integer.

    Args:
        n (int): The starting integer to find the next prime number.

    Returns:
        int: The next prime number greater than or equal to `n`.
    """
    n = int(n)
    if n < 2:
        return 2
    if n == 2:
        return 3
    if n % 6 == 0:
        if miller_rabin(n + 1):
            return n + 1
        n += 6
    elif 1 <= n % 6 < 5:
        n += 6 - n % 6
    else:
        if miller_rabin(n + 2):
            return n + 2
        n += 7
    while True:
        if miller_rabin(n - 1):
            return n - 1
        if miller_rabin(n + 1):
            return n + 1
        n += 6


def prevprime(n: int, raise_error: bool = True) -> int:
    """
    Find the largest prime number less than the given integer `n`.

    Notes:
        - The function uses the Miller-Rabin primality test to check for prime numbers.
        - The algorithm optimizes the search by leveraging properties of numbers
          in the 6k ± 1 form, where k is an integer.

    Args:
        n (int): The integer to find the previous prime number for.
        raise_error (bool, optional): If True, raises a ValueError when no prime
            number exists below `n`. Defaults to True.

    Returns:
        int: The largest prime number less than `n`. If no such prime exists
            and `raise_error` is False, returns 0.

    Raises:
        ValueError: If `n` is less than or equal to 2 and `raise_error` is True.
    """
    n = int(n)
    if n <= 2:
        if raise_error:
            raise ValueError(f"No prime number less than {n}")
        return 0
    if n == 3:
        return 2
    if n <= 5:
        return 3
    if n % 6 == 0:
        if miller_rabin(n - 1):
            return n - 1
        n -= 6
    elif n % 6 == 1:
        if miller_rabin(n - 2):
            return n - 2
        n -= 7
    else:
        n -= n % 6
    while True:
        if miller_rabin(n + 1):
            return n + 1
        if miller_rabin(n - 1):
            return n - 1
        n -= 6


def randprime(a: int, b: int, raise_error: bool = True) -> int:
    """
    Generate a random prime number within the specified range [a, b].

    This function attempts to find a random prime number between the integers
    `a` and `b` (inclusive). If no prime number exists in the range, it either
    raises a ValueError or returns 0, depending on the `raise_error` flag.

    Note: This function cannot uniformly randomly select from all primes in the range.

    Args:
        a(int): The lower bound of the range (inclusive).
        b(int): The upper bound of the range (inclusive).
        raise_error (bool, optional): If True, raises a ValueError when no prime
            number is found in the range. If False, returns 0 in such cases.
            Defaults to True.

    Returns:
        int: A prime number within the range [a, b], or 0 if no prime is found
            and `raise_error` is set to False.

    Raises:
        ValueError: If no prime number exists in the range [a, b] and
            `raise_error` is True.
    """
    st = random.randint(a, b)
    if miller_rabin(st):
        return st
    nxt = nextprime(st)
    if nxt <= b:
        return nxt
    pre = prevprime(st)
    if pre >= a:
        return pre
    if raise_error:
        raise ValueError(f"No prime number in the range [{a}, {b}]")
    return 0



"""
This module provides utilities for creating and manipulating graphs.
"""

from typing import Any, Callable, Dict

from rdgc.graph.base import Graph
from rdgc.graph.utils import (
    null,
    complete,
    tournament,
    random_graph,
    cycle,
    wheel,
    union,
)
from rdgc.graph.tree import chain, star, tree, binary_tree, spanning_tree
from rdgc.graph.degree import from_degree_sequence, k_regular
from rdgc.graph.connected import connected, strongly_connected

__all__ = [
    "Graph",
    "make_graph",
    "GRAPH_GENERS",
    "null",
    "complete",
    "tournament",
    "random_graph",
    "chain",
    "star",
    "tree",
    "binary_tree",
    "spanning_tree",
    "cycle",
    "wheel",
    "union",
    "connected",
    "strongly_connected",
    "from_degree_sequence",
    "k_regular",
]

GRAPH_GENERS: Dict[str, Callable[..., Graph]] = {
    "null": null,
    "complete": complete,
    "tournament": tournament,
    "random": random_graph,
    "random_graph": random_graph,
    "chain": chain,
    "star": star,
    "tree": tree,
    "binary_tree": binary_tree,
    "spanning_tree": spanning_tree,
    "cycle": cycle,
    "wheel": wheel,
    "connected": connected,
    "strongly_connected": strongly_connected,
    "from_degree_sequence": from_degree_sequence,
    "k_regular": k_regular,
}


def make_graph(_type: str, *args: Any, **kwargs: Any) -> Graph:
    """
    Returns a graph of the specified type.

    Args:
        name (str): The type of graph.
        *args: The arguments to pass to the graph constructor.
        **kwargs: The keyword arguments to pass to the graph constructor.

    Raises:
        ValueError: If the graph type is unknown.

    Warnings:
        RuntimeWarning: If extra arguments are provided

    Returns:
        Graph: A graph of the specified type.
    """
    if _type not in GRAPH_GENERS:
        raise ValueError(f"Unknown graph type: {_type}")
    return GRAPH_GENERS[_type](*args, **kwargs)



"""
This module provides functions to generate tree graphs.
"""

import random
import warnings

from typing import Any, Callable, Optional, cast

from rdgc.utils import Dsu
from rdgc.graph.base import Graph


def chain(
    size: int,
    *args: Any,
    directed: bool = False,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    **kwargs: Any,
) -> "Graph":
    """
    Returns a chain graph with `size` vertices.

    Args:
        size (int): The number of vertices.
        directed (bool, optional): Specifies whether the graph is directed. Defaults to False.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A chain graph with `size` vertices.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    return tree(size, 1.0, 0.0, directed=directed, weight_gener=weight_gener)


def star(
    size: int,
    *args: Any,
    directed: bool = False,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    **kwargs: Any,
) -> "Graph":
    """
    Returns a star graph with `size` vertices.

    Args:
        size (int): The number of vertices.
        directed (bool, optional): Specifies whether the graph is directed. Defaults to False.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A star graph with `size` vertices.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    return tree(size, 0.0, 1.0, directed=directed, weight_gener=weight_gener)


def tree(
    size: int,
    chain_ratio: float = 0.0,
    star_ratio: float = 0.0,
    *args: Any,
    directed: bool = False,
    rank: bool = True,
    root_rank: Any = 0,
    chain_rank: Any = 1,
    star_rank: Any = 2,
    rand_rank: Any = -1,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    father_gener: Optional[Callable[[int], int]] = None,
    **kwargs: Any,
) -> "Graph":
    """
    Returns a tree graph with `size` vertices.

    Args:
        size (int): The number of vertices.
        chain (float, optional): The proportion of chain edges. Defaults to 0.0.
        star (float, optional): The proportion of star edges. Defaults to 0.0.
        directed (bool, optional): Specifies whether the graph is directed. Defaults to False.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.
        father_gener (Callable[[int], int], optional): A function to generate parent vertices. Defaults to None.

    Raises:
        ValueError: If the parameters are invalid.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A tree graph with `size` vertices.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    if chain_ratio + star_ratio > 1.0 or chain_ratio < 0.0 or star_ratio < 0.0:
        raise ValueError("Invalid parameters")

    if weight_gener is None:
        weight_gener = lambda u, v: None
    if father_gener is None:
        father_gener = lambda u: random.randint(0, u - 1)
    if not rank:
        root_rank = chain_rank = star_rank = rand_rank = None

    chain_cnt = int((size - 1) * chain_ratio)
    star_cnt = int((size - 1) * star_ratio)
    tree_graph = Graph(size, directed)
    tree_graph.set_rank(0, root_rank)
    for i in range(1, chain_cnt + 1):
        tree_graph.set_rank(i, chain_rank)
        tree_graph.add_edge(i, i - 1, weight_gener(i, i - 1))
    for i in range(chain_cnt + 1, chain_cnt + star_cnt + 1):
        tree_graph.set_rank(i, star_rank)
        tree_graph.add_edge(i, chain_cnt, weight_gener(i, chain_cnt))
    for i in range(chain_cnt + star_cnt + 1, size):
        father = father_gener(i)
        tree_graph.set_rank(i, rand_rank)
        tree_graph.add_edge(i, father, weight_gener(i, father))
    return tree_graph


def binary_tree(
    size: int,
    left: Optional[float] = None,
    right: Optional[float] = None,
    *args: Any,
    root: int = 0,
    rank: bool = True,
    root_rank: Any = 0,
    left_rank: Any = -1,
    right_rank: Any = 1,
    directed: bool = False,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    **kwargs: Any,
) -> "Graph":
    """
    Returns a binary tree graph with `size` vertices.

    Args:
        size (int): The number of vertices.
        left (float, optional): The proportion of left edges. Defaults to None.
        right (float, optional): The proportion of right edges. Defaults to None.
        root (int, optional): The root vertex. Defaults to 0.
        root_rank (Any, optional): The rank of the root vertex. Defaults to 0.
        left_rank (Any, optional): The rank of the left child. Defaults to -1.
        right_rank (Any, optional): The rank of the right child. Defaults to 1.
        directed (bool, optional): Specifies whether the graph is directed. Defaults to False.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.

    Raises:
        ValueError: If the root vertex is invalid.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A binary tree graph with `size` vertices.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    if root < 0 or root >= size:
        raise ValueError("Invalid root")

    if weight_gener is None:
        weight_gener = lambda u, v: None
    if not rank:
        root_rank = left_rank = right_rank = None
    if left is None and right is None:
        rnk = 0.5
    elif left is None:
        rnk = 1.0 - cast(float, right)
    elif right is None:
        rnk = left
    else:
        rnk = left / (left + right)

    btree = Graph(size, directed)
    left_rt, right_rt = [root], [root]
    for i in range(size):
        if i == root:
            btree.set_rank(i, root_rank)
            continue
        if random.random() < rnk:
            btree.set_rank(i, left_rank)
            rt = random.randrange(len(left_rt))
            btree.add_edge(i, left_rt[rt], weight_gener(i, left_rt[rt]))
            left_rt[rt] = i
            right_rt.append(i)
        else:
            btree.set_rank(i, right_rank)
            rt = random.randrange(len(right_rt))
            btree.add_edge(i, right_rt[rt], weight_gener(i, right_rt[rt]))
            right_rt[rt] = i
            left_rt.append(i)
    return btree


def spanning_tree(
    size: int,
    *args: Any,
    weight_gener: Optional[Callable[[int, int], Any]] = None,
    **kwargs: Any,
) -> "Graph":
    """
    Returns a tree graph with `size` vertices, which is generated by union-find.

    Args:
        size (int): The number of vertices.
        weight_gener (Callable[[int, int], Any], optional): A function to generate edge weights. Defaults to None.

    Warnings:
        RuntimeWarning: If extra arguments are provided.

    Returns:
        Graph: A tree graph with `size` vertices.
    """
    if args or kwargs:
        warnings.warn("Extra arguments are ignored", RuntimeWarning, 2)
    if weight_gener is None:
        weight_gener = lambda u, v: None

    tree_graph = Graph(size)
    dsu_instance = Dsu(size)
    while tree_graph.edges < size - 1:
        u, v = random.sample(range(size), 2)
        if dsu_instance.union(u, v):
            tree_graph.add_edge(u, v, weight_gener(u, v))
    return tree_graph
